/*
 * caching_sequence_index.cpp
 *
 *  Created on: 2012/03/01
 *      Author: shu
 */
#include <fstream>
#include <vector>
#include <stdint.h>
#include <assert.h>
#include "encoder.h"
#include "common_parameters.h"
#include "cached_sa_sequence_index.h"

CachedSaSequenceIndex::CachedSaSequenceIndex(const unsigned char *text,
    unsigned int length, unsigned int max_cached_offset, const Encoder &encoder) :
    max_cached_offset_(max_cached_offset), max_character_(
        encoder.GetMaxRegularLetterCode()), key_shift_size_(0), cached_starts_(
        0), cached_ends_(0) {
  vector<Code> index_seq(length);
  Code min_regular_letter_code = encoder.GetMinRegularLetterCode();
  Code max_regular_letter_code = encoder.GetMaxRegularLetterCode();
  Code unknown_code = encoder.GetUnknownCode();
  int count = 0;
  for (uint32_t i = 0; i < index_seq.size(); ++i) {
    Code c = text[i];
    if (min_regular_letter_code <= c && c <= max_regular_letter_code) {
      index_seq[i] = c;
      count = 0;
    } else {
      if (count < 2 && c != CommonParameters::kSequenceEnd) {
        index_seq[i] = encoder.Encode('A');
        ++count;
      } else {
        index_seq[i] = unknown_code;
      }
    }
  }
  sa_.Rebuild(&index_seq[0], index_seq.size());
  SetKeyShiftSize(max_character_);
  SetCache(max_cached_offset_, max_character_);
}

CachedSaSequenceIndex::CachedSaSequenceIndex(ifstream &is) :
    max_cached_offset_(0), max_character_(0), key_shift_size_(0), cached_starts_(
        0), cached_ends_(0) {
  Load(is);
}

CachedSaSequenceIndex::~CachedSaSequenceIndex() {
}

void CachedSaSequenceIndex::Search(unsigned int offset, unsigned int start,
    unsigned int end, unsigned char *start_character,
    unsigned char *end_character, unsigned int *start_list,
    unsigned int *end_list) {
  if (offset <= max_cached_offset_) {
    if (sa_.Locate(start) + offset >= sa_.GetTextSize()) {
      ++start;
    }
    if (start > end) {
      *end_character = *start_character;
      start_list[*start_character] = 1;
      end_list[*start_character] = 0;
      return;
    }

    unsigned int start_position = sa_.Locate(start);
    unsigned int start_key = GetKey(start_position, offset) << key_shift_size_;
    unsigned char s_c = sa_.GetText()[start_position + offset];
    unsigned char e_c = s_c;
    if (e_c > *end_character) {
      *end_character = *start_character;
      start_list[*start_character] = 1;
      end_list[*start_character] = 0;
      return;
    }

    for (unsigned char c = s_c; c <= max_character_; ++c) {
      unsigned int key = start_key + c;
      start_list[c] = cached_starts_[offset][key];
      end_list[c] = cached_ends_[offset][key];
      if (start_list[c] <= end_list[c]) {
        e_c = c;
      }
    }
    *start_character = s_c;
    *end_character = e_c;
  } else {
    sa_.Search(offset, start, end, start_character, end_character, start_list,
        end_list);
  }
}

unsigned int CachedSaSequenceIndex::Locate(unsigned int position) {
  return sa_.Locate(position);
}

unsigned int CachedSaSequenceIndex::GetMaxAlphabet() {
  return sa_.GetMaxAlphabet();
}

void CachedSaSequenceIndex::Load(ifstream &is) {
  sa_.Load(is);
  is.read((char*) (&max_cached_offset_), sizeof(max_cached_offset_));
  is.read((char*) (&key_shift_size_), sizeof(key_shift_size_));
  is.read((char*) (&max_character_), sizeof(max_character_));
  cached_starts_.resize(max_cached_offset_ + 1);
  cached_ends_.resize(max_cached_offset_ + 1);

  for (unsigned int i = 0; i <= max_cached_offset_; ++i) {
    size_t cache_size = 0;
    is.read((char*) (&cache_size), sizeof(cache_size));
    cached_starts_[i].resize(cache_size);
    cached_ends_[i].resize(cache_size);
    is.read((char*) (&cached_starts_[i][0]), cache_size);
    is.read((char*) (&cached_ends_[i][0]), cache_size);
  }
}
void CachedSaSequenceIndex::Save(ofstream &os) {
  sa_.Save(os);
  os.write((const char*) (&max_cached_offset_), sizeof(max_cached_offset_));
  os.write((const char*) (&key_shift_size_), sizeof(key_shift_size_));
  os.write((const char*) (&max_character_), sizeof(max_character_));
  for (unsigned int i = 0; i <= max_cached_offset_; ++i) {
    size_t cache_size = sizeof(cached_starts_[i][0]) * cached_starts_[i].size();
    os.write((const char*) (&cache_size), sizeof(cache_size));
    os.write((const char*) (&cached_starts_[i][0]), cache_size);
    os.write((const char*) (&cached_ends_[i][0]), cache_size);
  }
}

void CachedSaSequenceIndex::SetKeyShiftSize(unsigned char max_character) {
  unsigned int s = 0;
  unsigned char m = max_character;
  for (s = 0; m > 0; ++s, m >>= 1) {
  }
  key_shift_size_ = s;
}

unsigned int CachedSaSequenceIndex::GetKey(unsigned int position,
    unsigned int lenght) {
  const unsigned char *sa_text = sa_.GetText();
  unsigned int key = 0;
  unsigned int p_end = position + lenght;
  for (unsigned int p = position; p < p_end; ++p) {
    key <<= key_shift_size_;
    key += sa_text[p];
  }
  return key;

}

void CachedSaSequenceIndex::SetCache(unsigned int max_cached_offset,
    unsigned char max_character) {
  unsigned char s_c = 0;
  unsigned char e_c = max_character;
  vector<unsigned int> starts(max_character + 1);
  vector<unsigned int> ends(max_character + 1);

  cached_starts_.resize(max_cached_offset + 1);
  cached_ends_.resize(max_cached_offset + 1);

  for (unsigned int i = 0; i <= max_cached_offset; ++i) {
    unsigned int cache_size = 1 << (i + 1) * key_shift_size_;
    cached_starts_[i].resize(cache_size, 1);
    cached_ends_[i].resize(cache_size, 0);
  }

  sa_.Search(0, 0, sa_.GetTextSize() - 1, &s_c, &e_c, &starts[0], &ends[0]);
  for (unsigned char c = 0; c <= max_character; ++c) {
    if (s_c <= c && c <= e_c) {
      cached_starts_[0][c] = starts[c];
      cached_ends_[0][c] = ends[c];
    } else {
      cached_starts_[0][c] = 1;
      cached_ends_[0][c] = 0;
    }
  }

  unsigned int offset = 1;
  unsigned int key = 0;
  vector<unsigned char> c_list(max_cached_offset + 2, 0);

  while (offset > 0) {
    if (offset > max_cached_offset || c_list[offset] > max_character) {
      --offset;
      ++c_list[offset];
      key >>= 2 * key_shift_size_;
      key <<= key_shift_size_;
      if (c_list[offset] <= max_character) {
        key += c_list[offset];
      }
    } else {
      s_c = 0;
      e_c = max_character;
      unsigned int prev_offset = offset - 1;
      sa_.Search(offset, cached_starts_[prev_offset][key], cached_ends_[prev_offset][key], &s_c, &e_c, &starts[0], &ends[0]);
      unsigned int key_offset = key << key_shift_size_;
      for (int c = 0; c <= max_character; ++c) {
        unsigned int key = key_offset + c;
        if (s_c <= c && c <= e_c) {
          cached_starts_[offset][key] = starts[c];
          cached_ends_[offset][key] = ends[c];
        } else {
          cached_starts_[offset][key] = 1;
          cached_ends_[offset][key] = 0;
        }

      }
      ++offset;
      c_list[offset] = 0;
      key = key_offset;
    }
  }
}
